/* mswdynld - Dynamic loading and C function calling routines.         */
/* XLISP-STAT 2.1 Copyright (c) 1990, by Luke Tierney                  */
/* Additions to Xlisp 2.1, Copyright (c) 1989 by David Michael Betz    */
/* You may give out copies of this software; for conditions see the    */
/* file COPYING included with this distribution.                       */

/* Calling conventions are based on the conventions given in the New S */
/* book.                                                               */

#include "xlisp.h"

#include "xlsx.h"

#define nil 0L

#define seqlen(x) ((vectorp(x)) ? getsize(x) : llength(x))

extern LVAL mklist(), getnextelement();
extern double makedouble();

/************************************************************************/
/**                                                                    **/
/**                      DLL Handling Functions                        **/
/**                                                                    **/
/************************************************************************/

LVAL xsload_dll()
{
  char *dllname;
  HANDLE hdll;

  dllname = (char *) getstring(xlgastring());
  xllastarg();
  hdll = LoadLibrary(dllname);
  return((hdll >= 32) ? cvfixnum((FIXTYPE) hdll) : NIL);
}

LVAL xsfree_dll()
{
  HANDLE hdll;

  hdll = (HANDLE) getfixnum(xlgafixnum());
  xllastarg();
  if (hdll >= 32) FreeLibrary(hdll);
  return(NIL);
}

/************************************************************************/
/**                                                                    **/
/**               Allocation and Error Signalling Functions            **/
/**                                                                    **/
/************************************************************************/

static LVAL current_allocs;

/* allocate space that will be garbage collected after return */
static char *xscall_alloc(n, m)
     int n, m;
{
  LVAL adata;
  char *p;

  adata = newadata(n, m, FALSE);
  if (adata == NIL || (p = getadaddr(adata)) == nil)
    xlfail("allocation failed");
  current_allocs = cons(adata, current_allocs);
  return(p);
}

/************************************************************************/
/**                                                                    **/
/**                Lisp to C/FORTRAN Data Conversion                   **/
/**                                                                    **/
/************************************************************************/

#define IN 0
#define RE 1

typedef struct {
  int type, size;
  char *addr;
} call_arg;

/* convert lisp argument to allocated pointer */
static call_arg lisp2arg(x)
     LVAL x;
{
  call_arg a;
  LVAL elem, data;
  int i;

  xlprot1(x);

  /* make sure x is a sequence and find its length */
  if (! sequencep(x)) x = consa(x);
  a.size = seqlen(x);

  /* determine the mode of the data */
  for (i = 0, a.type = IN, data = x; i < a.size; i++) {
    elem = getnextelement(&data, i);
    if (floatp(elem)) a.type = RE;
    else if (! fixp(elem)) xlerror("not a real number", elem);
  }

  /* allocate space for the data */
  a.addr = xscall_alloc(a.size, (a.type == IN) ? sizeof(int) : sizeof(double));

  /* fill the space */
  for (i = 0, data = x; i < a.size; i++) {
    elem = getnextelement(&data, i);
    if (a.type == IN) ((int *) a.addr)[i] = getfixnum(elem);
    else ((double *) a.addr)[i] = makedouble(elem);
  }
  
  xlpop();
  return(a);
}

/* copy allocated pointer back to new lisp list */
static LVAL arg2lisp(a)
     call_arg a;
{
  LVAL x, next;
  int i;

  xlsave1(x);
  x = mklist(a.size, NIL);
  for (i = 0, next = x; i < a.size; i++, next = cdr(next)) {
    if (a.type == IN) rplaca(next, cvfixnum((FIXTYPE) ((int *) a.addr)[i]));
    else rplaca(next, cvflonum((FLOTYPE) ((double *) a.addr)[i]));
  }
  xlpop();
  return(x);
}

/************************************************************************/
/**                                                                    **/
/**                 Foreign Function Call Function                     **/
/**                                                                    **/
/************************************************************************/

LVAL xscall_cfun()
{
  LVAL result, Lfun, old_allocs, next;
  call_arg *args, *pargs;
  int nargs, i;
  FARPROC routine;
  HANDLE hdll;
  char *name;
  XLSXblock params;

  // ### patch this to handle errors properly (reset allocs, etc)
  xlstkcheck(3);
  xlsave(old_allocs);
  xlprotect(current_allocs);
  xlsave(result);
  old_allocs = current_allocs;
  current_allocs = NIL;

  /* get the routine pointer */
  hdll = (HANDLE) getfixnum(xlgafixnum());
  Lfun = xlgetarg();
  if (stringp(Lfun)) name = (char *) getstring(Lfun);
  else if (fixp(Lfun)) name = MAKEINTRESOURCE((int) getfixnum(Lfun));
  routine = GetProcAddress(hdll, name);
  if (! routine) xlerror("can't find function address", Lfun);

  /* convert the arguments to allocated pointers */
  nargs = xlargc;
  if (nargs == 0) xlfail("too few arguments");
  args = (call_arg *) xscall_alloc(nargs, sizeof(call_arg));
  params.argc = nargs;
  params.argv = (char **) xscall_alloc(nargs, sizeof(char *));
  for (i = 0; i < nargs; i++) {
    args[i] = lisp2arg(xlgetarg());
	params.argv[i] = args[i].addr;
  }

  /* make the call */
  (*routine)(&params);

  /* convert the pointers back to lists, grouped in a list */
  result = (nargs > 0) ? mklist(nargs, NIL) : NIL;
  for (next = result, pargs = args; consp(next); next = cdr(next), pargs++)
    rplaca(next, arg2lisp(*pargs));
  
  current_allocs = old_allocs;
  xlpopn(3);

  return(result);
}
